@@ -160,21 +160,25 @@ const MappingDesc kMemoryLayout[] = {
 # define SHADOW_TO_ORIGIN(shadow) (((uptr)(shadow)) + 0x1000000000ULL)
 
 #elif SANITIZER_LINUX && SANITIZER_PPC64
-
 const MappingDesc kMemoryLayout[] = {
-    {0x000000000000ULL, 0x000100000000ULL, MappingDesc::APP, "low memory"},
-    {0x000100000000ULL, 0x080000000000ULL, MappingDesc::INVALID, "invalid"},
-    {0x080000000000ULL, 0x180100000000ULL, MappingDesc::SHADOW, "shadow"},
-    {0x180100000000ULL, 0x1C0000000000ULL, MappingDesc::INVALID, "invalid"},
-    {0x1C0000000000ULL, 0x2C0100000000ULL, MappingDesc::ORIGIN, "origin"},
-    {0x2C0100000000ULL, 0x300000000000ULL, MappingDesc::INVALID, "invalid"},
-    {0x300000000000ULL, 0x400000000000ULL, MappingDesc::APP, "high memory"}};
+    {0x000000000000ULL, 0x000200000000ULL, MappingDesc::APP, "low memory"},
+    {0x000200000000ULL, 0x080000000000ULL, MappingDesc::INVALID, "invalid"},
+    {0x080000000000ULL, 0x180200000000ULL, MappingDesc::SHADOW, "shadow"},
+    {0x180200000000ULL, 0x1C0000000000ULL, MappingDesc::INVALID, "invalid"},
+    {0x1C0000000000ULL, 0x2C0200000000ULL, MappingDesc::ORIGIN, "origin"},
+    {0x2C0200000000ULL, 0x300000000000ULL, MappingDesc::INVALID, "invalid"},
+    {0x300000000000ULL, 0x800000000000ULL, MappingDesc::APP, "high memory"}};
 
+// Various kernels use different low end ranges but we can combine them into one
+// big range. They also use different high end ranges but we can map them all to
+// one range.
 // Maps low and high app ranges to contiguous space with zero base:
-//   Low:  0000 0000 0000 - 0000 ffff ffff  ->  1000 0000 0000 - 1000 ffff ffff
+//   Low:  0000 0000 0000 - 0001 ffff ffff  ->  1000 0000 0000 - 1001 ffff ffff
 //   High: 3000 0000 0000 - 3fff ffff ffff  ->  0000 0000 0000 - 0fff ffff ffff
+//   High: 4000 0000 0000 - 4fff ffff ffff  ->  0000 0000 0000 - 0fff ffff ffff
+//   High: 7000 0000 0000 - 7fff ffff ffff  ->  0000 0000 0000 - 0fff ffff ffff
 #define LINEARIZE_MEM(mem) \
-  (((uptr)(mem) & ~0x200000000000ULL) ^ 0x100000000000ULL)
+  (((uptr)(mem) & ~0xE00000000000ULL) ^ 0x100000000000ULL)
 #define MEM_TO_SHADOW(mem) (LINEARIZE_MEM((mem)) + 0x080000000000ULL)
 #define SHADOW_TO_ORIGIN(shadow) (((uptr)(shadow)) + 0x140000000000ULL)
 
@@ -199,7 +203,7 @@ const MappingDesc kMemoryLayout[] = {
 #define MEM_TO_SHADOW(mem) (LINEARIZE_MEM((mem)) + 0x100000000000ULL)
 #define SHADOW_TO_ORIGIN(shadow) (((uptr)(shadow)) + 0x280000000000)
 
-#elif SANITIZER_LINUX && SANITIZER_WORDSIZE == 64
+#elif SANITIZER_NETBSD || (SANITIZER_LINUX && SANITIZER_WORDSIZE == 64)
 
 #ifdef MSAN_LINUX_X86_64_OLD_MAPPING
 // Requries PIE binary and ASLR enabled.
@@ -310,7 +314,7 @@ void PrintWarning(uptr pc, uptr bp);
 void PrintWarningWithOrigin(uptr pc, uptr bp, u32 origin);
 
 void GetStackTrace(BufferedStackTrace *stack, uptr max_s, uptr pc, uptr bp,
-                   bool request_fast_unwind);
+                   void *context, bool request_fast_unwind);
 
 void ReportUMR(StackTrace *stack, u32 origin);
 void ReportExpectedUMRNotFound(StackTrace *stack);
@@ -330,32 +334,32 @@ u32 ChainOrigin(u32 id, StackTrace *stack);
 
 const int STACK_TRACE_TAG_POISON = StackTrace::TAG_CUSTOM + 1;
 
-#define GET_MALLOC_STACK_TRACE                                                 \
-  BufferedStackTrace stack;                                                    \
-  if (__msan_get_track_origins() && msan_inited)                               \
-  GetStackTrace(&stack, common_flags()->malloc_context_size,                   \
-                StackTrace::GetCurrentPc(), GET_CURRENT_FRAME(),               \
+#define GET_MALLOC_STACK_TRACE                                            \
+  BufferedStackTrace stack;                                               \
+  if (__msan_get_track_origins() && msan_inited)                          \
+  GetStackTrace(&stack, common_flags()->malloc_context_size,              \
+                StackTrace::GetCurrentPc(), GET_CURRENT_FRAME(), nullptr, \
                 common_flags()->fast_unwind_on_malloc)
 
 // For platforms which support slow unwinder only, we restrict the store context
 // size to 1, basically only storing the current pc. We do this because the slow
 // unwinder which is based on libunwind is not async signal safe and causes
 // random freezes in forking applications as well as in signal handlers.
-#define GET_STORE_STACK_TRACE_PC_BP(pc, bp)                                    \
-  BufferedStackTrace stack;                                                    \
-  if (__msan_get_track_origins() > 1 && msan_inited) {                         \
-    if (!SANITIZER_CAN_FAST_UNWIND)                                            \
-      GetStackTrace(&stack, Min(1, flags()->store_context_size), pc, bp,       \
-                    false);                                                    \
-    else                                                                       \
-      GetStackTrace(&stack, flags()->store_context_size, pc, bp,               \
-                    common_flags()->fast_unwind_on_malloc);                    \
+#define GET_STORE_STACK_TRACE_PC_BP(pc, bp)                               \
+  BufferedStackTrace stack;                                               \
+  if (__msan_get_track_origins() > 1 && msan_inited) {                    \
+    if (!SANITIZER_CAN_FAST_UNWIND)                                       \
+      GetStackTrace(&stack, Min(1, flags()->store_context_size), pc, bp,  \
+                    nullptr, false);                                      \
+    else                                                                  \
+      GetStackTrace(&stack, flags()->store_context_size, pc, bp, nullptr, \
+                    common_flags()->fast_unwind_on_malloc);               \
   }
 
-#define GET_FATAL_STACK_TRACE_PC_BP(pc, bp)                                    \
-  BufferedStackTrace stack;                                                    \
-  if (msan_inited)                                                             \
-  GetStackTrace(&stack, kStackTraceMax, pc, bp,                                \
+#define GET_FATAL_STACK_TRACE_PC_BP(pc, bp)              \
+  BufferedStackTrace stack;                              \
+  if (msan_inited)                                       \
+  GetStackTrace(&stack, kStackTraceMax, pc, bp, nullptr, \
                 common_flags()->fast_unwind_on_fatal)
 
 #define GET_STORE_STACK_TRACE \
